"""
The implementation.
"""

import logging
from shutil import rmtree
from subprocess import run


_log = logging.getLogger(__name__)


class Vfs(object):
    """An abstract filesystem device which can accumulate changes, and apply them in a batch."""

    def __init__(self, log=None):
        self._log = log or []

    def execute(self, /, callback=None):
        for e in self._log:
            _log.debug(e)
            if callback:
                callback(e)

            if e[0] == "exec":
                _, dir, cmd = e
                run(cmd, cwd=str(dir))

            elif e[0] == "link":
                _, src, dest = e
                if dest.is_file() or dest.is_symlink():
                    if dest.is_symlink() and dest.readlink() == src:
                        continue
                    else:
                        _log.warn(f"Replacing {dest}")
                        dest.unlink()
                elif dest.is_dir():
                    _log.warn(f"Replacing {dest}")
                    rmtree(dest)

                assert not dest.exists(), f"{dest} should not exist"
                dest.symlink_to(src)

            elif e[0] == "copy":
                raise NotImplementedError()

            elif e[0] == "chmod":
                _, dest, mode = e
                dest.chmod(mode)

            elif e[0] == "mkdir":
                _, dest = e
                if dest.is_dir():
                    continue
                elif dest.exists() or dest.is_symlink():
                    dest.unlink()

                dest.mkdir(exist_ok=True)

            elif e[0] == "unlink":
                _, dest = e
                # Note that a path which is a dangling symlink will NOT exist but WILL be a symlink
                if not dest.exists() and not dest.is_symlink():
                    continue
                # Files and dirs just unlink
                if dest.is_symlink() or dest.is_file():
                    dest.unlink()
                # Dirs require recursion
                elif dest.is_dir():
                    rmtree(dest)
                # Don't succeed silently
                else:
                    raise Exception(f"Couldn't unlink {dest}")

    def _append(self, msg):
        self._log.append(msg)

    def link(self, src, dest):
        self._append(("link", src, dest))

    def copy(self, src, dest):
        self._append(("copy", src, dest))

    def chmod(self, dest, mode):
        self._append(("chmod", dest, mode))

    def mkdir(self, dest):
        self._append(("mkdir", dest))

    def exec(self, dest, cmd):
        self._append(("exec", dest, cmd))

    def unlink(self, dest):
        self._append(("unlink", dest))

    def copy(self):
        return Vfs(list(self._log))

    def merge(self, other: "Vfs"):
        self._log.extend(other._log)
